########################################################################
#
#	Author: Hung Jung Lu <hungjunglu@hotmail.com>
#
# ======================================================================
#
#  Disclaimer
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions
#   are met:
#
#     Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#
#     The name of the author may not be used to endorse or promote 
#     products derived from this software without specific prior
#     written permission
#
#   THIS SOFTWARE IS PROVIDED BY HUNG JUNG LU ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL HUNG JUNG LU OR HIS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
#
########################################################################

__version__='0.1.1'

from Acquisition import Implicit
from Globals import HTMLFile
from Globals import Persistent
from Shared.DC.ZRDB.TM import TM
from UserDict import UserDict

import AccessControl.Role
import OFS.SimpleItem 
import cPickle 
import string
import thread
import time
import whrandom

class SessionNotFoundError(LookupError):
    pass
    
class ServerSessionData:
    """String with a time stamp"""
    
    def __init__(self, data):
        self.timestamp = time.time()
        self.data = data

    def getData(self):
        self.timestamp = time.time()
        return self.data

    def setData(self, data):
        self.timestamp = time.time()
        self.data = data

    def getTimeStamp(self):
        return self.timestamp

    def setTimeStamp(self):
        self.timestamp = time.time()
        
class HappySessionServer_DefaultCoreServer:
    """Implements a RAM-based session server"""
    #
    # We use client-server design because in the future we may
    # want to use a single session server to serve several copies
    # of Zope. This also allows future implementation of different
    # types of servers, based on RAM, File System, Database, etc.
    #
    def __init__(self, cookieName='SessionUID', maxIdle=3600.):
        self._cookieName = cookieName
        self._cookieLength = 128
        self._sessions = {}
        self._maxIdle = maxIdle
        self._alphanums = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789'
        self._lock = thread.allocate_lock()
    
    def _checkUpdateExpiration(self, SUID):
        if self._sessions.has_key(SUID):
            t = self._sessions[SUID].getTimeStamp()
            if time.time() - t > self._maxIdle:
                del self._sessions[SUID]
            else:
                self._sessions[SUID].setTimeStamp()

    def _randomExpiration(self):
        if len(self._sessions) > 0: # randomly check expired sessions
            SUID = whrandom.choice(self._sessions.keys())
            t = self._sessions[SUID].getTimeStamp()
            if time.time() - t > self._maxIdle:
                del self._sessions[SUID]

    def _hasSession(self, SUID):
        self._checkUpdateExpiration(SUID)
        self._randomExpiration()
        return self._sessions.has_key(SUID)
                
    def getCookieName(self):
        return self._cookieName

    def getMaxIdle(self):
        return self._maxIdle
        
    def getNumberSessions(self):
        return len(self._sessions)

    def createSession(self):
        self._lock.acquire()
        chars = []
        for i in range(self._cookieLength):
            chars.append(whrandom.choice(self._alphanums))
        SUID = string.join(chars, '')
        if self._sessions.has_key(SUID):
            result = self.createSession() # recurse until a valid SUID found.
        else:
            self._sessions[SUID] = ServerSessionData('(d.')
            result = SUID
        self._lock.release() 
        return result
                    
    def retrieve(self, SUID):
        self._lock.acquire()
        if self._hasSession(SUID):
            result = self._sessions[SUID].getData()
        else:
            result = SessionNotFoundError()
        self._lock.release() 
        return result

    def store(self, SUID, data):
        self._lock.acquire()
        if self._hasSession(SUID):
            self._sessions[SUID].setData(data)
            result = None
        else:
            result = SessionNotFoundError()
        self._lock.release() 
        return result

class TransactionController(Implicit, TM):
    """Implements an object that is transaction aware"""

    def _begin(self,*ignored):
        self.aq_parent._onTransactionStarted()

    def _abort(self,*ignored):
        # do nothing
        pass
    
    def _finish(self,*ignored):
        self.aq_parent._onTransactionEnded()

class VirtualServerFile:
    """Adapter to communicate with the session server"""

    def __init__(self, server, SUID):
        self.server = server
        self.SUID = SUID
        self.data = ''
        self.position = 0
       
    def retrieve_from_server(self):
         self.data = self.server.retrieve(self.SUID)
         return self.data
            
    def store_to_server(self):
         result = self.server.store(self.SUID, self.data)
         return result
            
    def read(self, n=None):
        if n == None:
            result = self.data
            self.position = len(self.data)
        else:
            if self.position + n <= len(self.data):
                result = self.data[self.position:self.position+n]
                self.position = self.position + n
            else:
                result = self.data[self.position:]
                self.position = len(self.data)
        return result
        
    def readline(self):
        newposition = string.find(self.data, '\n', self.position) + 1
        if newposition == 0:
            result = self.data[self.position:]
            self.position = len(self.data)
        else:
            result = self.data[self.position:newposition]
            self.position = newposition
        return result
    
    def write(self, data):
        self.data = self.data + data
        
class HappySession(OFS.SimpleItem.Item, Persistent, Implicit, 
                   AccessControl.Role.RoleManager):
    """A session management client"""
    #
    # All attributes starting with _v_ are volatile. Volatile
    # attributes are meaningful only during the same REQUEST
    # transaction, and therefore must be set per each new REQUEST.
    #

    meta_type='HappySession'
    icon     ='misc_/HappySession/happysession'

    manage_options=(
    {'label':'Contents',   'action':'manage_main'},
    {'label':'Security',   'action':'manage_access'},
    )

    __ac_permissions__=(
    ('View management screens',
     ['manage','manage_menu','manage_main','manage_tabs' ]),
    )

    manage_main=HTMLFile('main', globals())
    
    servers = {}

    def __init__(self, id, title, cookie_name, max_idle):
        self.id = id
        self.title = title
        self.cookie_name = cookie_name # this should actually be server feature
        self.max_idle = max_idle # this should actually be server feature
        self.client_id = time.time() # use the creation time as the client id

    def _retrieve(self): 
        self._v_SUID = self.REQUEST.cookies[self.cookie_name]
        vsf = VirtualServerFile(self._v_server, self._v_SUID)
        result = vsf.retrieve_from_server()
        if isinstance(result, SessionNotFoundError):
            self._raiseSessionNotFound()
        self._v_data = cPickle.load(vsf)

    def _store(self):
        self._v_SUID = self.REQUEST.cookies[self.cookie_name]
        vsf = VirtualServerFile(self._v_server, self._v_SUID)
        cPickle.dump(self._v_data, vsf)
        result = vsf.store_to_server()
        if result == SessionNotFoundError:
            self._raiseSessionNotFound()

    def _raiseSessionNotFound(self):
        del self.REQUEST.cookies[self.cookie_name]
        #self.REQUEST.RESPONSE.expireCookie(self.cookie_name)
        self.REQUEST.RESPONSE.setCookie(self.cookie_name, '', path='/')        
        raise SessionNotFoundError('Session Not Found', 'Your sesssion has expired.')

    def _initialize(self):
        # initialize server
        if not hasattr(self, '_v_server'):
            if not HappySession.servers.has_key(self.client_id):
                HappySession.servers[self.client_id] = HappySessionServer_DefaultCoreServer(self.cookie_name, self.max_idle)
            self._v_server = HappySession.servers[self.client_id]
         # initialize transaction controller
        if not hasattr(self, '_v_transaction_controller'):
            self._v_transaction_controller = TransactionController()
        self._v_first_registration = 0
        self._v_transaction_controller._register()
        if self._v_first_registration:      
            if (not self.REQUEST.cookies.has_key(self.cookie_name) or
                self.REQUEST.cookies[self.cookie_name] == ''):
                self._v_SUID = self._v_server.createSession()
                self.REQUEST.cookies[self.cookie_name] = self._v_SUID
                self.REQUEST.RESPONSE.setCookie(self.cookie_name, self._v_SUID, path='/')
            self._retrieve()
            
    def _onTransactionStarted(self):
        self._v_first_registration = 1
        
    def _onTransactionEnded(self):
        self._store()

    def getNumberSessions(self):
        if HappySession.servers.has_key(self.client_id):
            return HappySession.servers[self.client_id].getNumberSessions()
        else:
            return 0
                        
    ## Use _v_data in dict methods

    def __repr__(self): 
        self._initialize()
        return repr(self._v_data)
        
    def __cmp__(self, dict):
        self._initialize()
        if isinstance(dict, UserDict):
            return cmp(self._v_data, dict._v_data)
        else:
            return cmp(self._v_data, dict)

    def __len__(self): 
        self._initialize()
        return len(self._v_data)
        
    def __getitem__(self, key):
        self._initialize()
        return self._v_data.get(key)
            
    def __setitem__(self, key, item):
        self._initialize()
        if item == None:
            if self._v_data.has_key(key):
                del self._v_data[key]
        else:
            self._v_data[key] = item
            
    def __delitem__(self, key): 
        self._initialize()
        del self._v_data[key]

    def clear(self): 
        self._initialize()
        self._v_data.clear()

    def keys(self): 
        self._initialize()
        return self._v_data.keys()

    def items(self): 
        self._initialize()
        return self._v_data.items()
        
    def values(self): 
        self._initialize()
        return self._v_data.values()
        
    def has_key(self, key): 
        self._initialize()
        return self._v_data.has_key(key)
        
    def update(self, dict):
        self._initialize()
        if isinstance(dict, UserDict):
            self._v_data.update(dict.data)
        elif isinstance(dict, type(self._v_data)):
            self._v_data.update(dict)
        else:
            for k, v in dict.items():
                self._v_data[k] = v
        for k in self._v_data.keys():
            if self._v_data[k] == None:
                del self._v_data[k]
                
    def copy(self):
        self._initialize()
        # copy doesn't make much sense in this context
        raise AttributeError,'Copy not supported for Session storage objects'

    set=__setitem__
       
manage_addHappySessionForm=HTMLFile('addHappySession', globals())

def manage_addHappySession(self, cookie_name, max_idle, REQUEST=None):
    """Method to add HappySession to a folder"""

    id = 'SESSION'
    title = 'HappySession'
    ob = HappySession(id, title, cookie_name, float(max_idle))
    self._setObject(id, ob)
    if REQUEST:
        return self.manage_main(self, REQUEST)

